from __future__ import annotations
from typing import Optional, Union, List, Tuple, Generic, TypeVar
from enum import Enum
from dataclasses import dataclass
from functools import reduce

import numpy as np

from . import plot

# A Mask is a 2D array representing the class index of each pixel of an image as values
Mask = List[List[int]]

# A Image is a 3D array representing RGB values or a 2D array representing grayscale values
Image = Union[List[List[List[int]]], List[List[int]]]

TResult = TypeVar('TResult')


@dataclass
class Result(Generic[TResult]):
    """Result of a GT versus Predictions

    Args:
        Generic (Result type): Type of the results.

    Attributes:
        image (Image): Image that the GT and predictions are based. Format: (Rows, Cols, Chanels) or (Rows, Cols) for grayscale.
        gt (TResult): Ground truth results.
        prediction (Optional[TResult]): Prediction results. None if only GT must be shown.
    """
    image: Image
    gt: TResult
    prediction: Optional[TResult] = None


class PredictionResultType(Enum):
    TP = 0
    FP = 1
    TN = 2
    FN = 3


@dataclass
class PredictionResult:
    score: float
    type: PredictionResultType


@dataclass
class BBox:
    """Bounding Box data structure

    Attributes:
        upper_left_x (int): Uper left X position of the Bounding Box.
        upper_left_y (int): Upper left Y position of the Bounding Box.
        width (int): Width of the Bounding Box.
        height (int): Height of the Bounding Box.
        cls (int): Bounding Box class index.
        score (Optional[float]): Bounding Box prediction score.
    """
    upper_left_x: int
    upper_left_y: int
    width: int
    height: int
    cls: int
    score: Optional[float] = None

    @property
    def upper_left_point(self) -> Tuple[int, int]:
        """Tuple[int, int]: (X,Y) of the upper left point of the Bounding Box."""
        return (self.upper_left_x, self.upper_left_y)

    @property
    def bottom_right_point(self) -> Tuple[int, int]:
        """Tuple[int, int]: (X,Y) of the bottom right point of the Bounding Box."""
        return (self.upper_left_x + self.width - 1, self.upper_left_y + self.height - 1)

    @property
    def center_point(self) -> Tuple[int, int]:
        """Tuple[int, int]: (X,Y) of the center point of the Bounding Box."""
        return ((self.upper_left_x + self.width - 1) // 2, (self.upper_left_y + self.height - 1) // 2)

    @property
    def area(self) -> int:
        """int: Area of the Bounding Box."""
        return self.width * self.height

    def intersection_area_with(self: BBox, bbox: BBox) -> int:
        """Calculates the intersection area with another bbox

        Args:
            self (BBox): This bbox
            bbox (BBox): Bbox to intersect with

        Returns:
            int: The intersection area with the bbox
        """

        # Gets each box upper left and bottom right coordinates
        (upr_lft_x_a, upr_lft_y_a) = self.upper_left_point
        (btm_rgt_x_a, btm_rgt_y_a) = self.bottom_right_point

        (upr_lft_x_b, upr_lft_y_b) = bbox.upper_left_point
        (btm_rgt_x_b, btm_rgt_y_b) = bbox.bottom_right_point

        # Calculates the intersection box upper left and bottom right coordinates
        (upr_lft_x_intersect, upr_lft_y_intersect) = (
            max(upr_lft_x_a, upr_lft_x_b), max(upr_lft_y_a, upr_lft_y_b))
        (btm_rgt_x_intersect, btm_rgt_y_intersect) = (
            min(btm_rgt_x_a, btm_rgt_x_b), min(btm_rgt_y_a, btm_rgt_y_b))

        # Calculates the height and width of the intersection box
        (w_intersect, h_intersect) = (btm_rgt_x_intersect -
                                      upr_lft_x_intersect + 1, btm_rgt_y_intersect - upr_lft_y_intersect + 1)

        # If H or W <= 0, there is no intersection
        if (w_intersect <= 0) or (h_intersect <= 0):
            return 0

        return w_intersect * h_intersect

    def union_area_with(self: BBox, bbox: BBox, intersection_area: Optional[int] = None) -> int:
        """Calculates the union area with another bbox

        Args:
            self (BBox): This bbox
            bbox (BBox): Bbox to union with
            intersection_area (Optional[int], optional): The intersection area between this and bbox. Defaults to None.

        Returns:
            int: The union area with the bbox
        """

        return self.area + bbox.area - (intersection_area or self.intersection_area_with(bbox))


@dataclass
class Classification:
    """Represents a classification

    Attributes:
        cls (int): Class index.
        score (Optional[float], optional): Prediction score. Defaults to None.
    """
    cls: int
    score: Optional[float] = None


@dataclass
class BinaryClassificationMetrics:
    """Binary classification metrics

    Attributes:
        cls (str): Class name.
        TP (int): True Positive count. Defaults to 0.
        FP (int): False Positive count. Defaults to 0.
        TN (int): True Negative count. Defaults to 0.
        FN (int): False Negative count. Defaults to 0.
    """

    cls: str
    TP: int = 0
    FP: int = 0
    TN: int = 0
    FN: int = 0

    @property
    def has_instances(self) -> bool:
        """int: Indicates if the class has any ground truth or predicted instances."""
        return self.count > 0

    @property
    def count(self) -> int:
        """int: Total count of classified instances."""
        return self.TP + self.TN + self.FP + self.FN

    @property
    def accuracy(self) -> float:
        """int: Total count of classified instances."""
        return (self.TP + self.TN)/self.count

    @property
    def recall(self) -> float:
        """float: Recall metric - TP / (TP + FN)."""
        return self.TP/(self.TP + self.FN)

    @property
    def false_positive_rate(self) -> float:
        """float: False Positive Rate (FPR) metric - FP / (FP + TN)."""
        return self.FP/(self.FP + self.TN)

    @property
    def specificity(self) -> float:
        """float: Specificity metric - TN / (FP + TN)."""
        return self.TN/(self.FP + self.TN)

    @property
    def precision(self) -> float:
        """float: Precision metric - TP / (FP + TP)."""
        return self.TP/(self.FP + self.TP)

    @property
    def f_score(self) -> float:
        """float: F-Score/Dice metric - 2*TP / (FP + FN + 2*TP)."""
        return 2*self.TP/(self.FP + self.FN + 2*self.TP)

    @property
    def confusion_matrix(self) -> List[List[int]]:
        """List[List[int]]: Confusion matrix of all the classes"""
        return [[self.TP, self.FP], [self.FN, self.TN]]

    def show_confusion_matrix(self):
        """Plots de confusion matrix

        Return:
            Tuple[Figure, Axes]: Figure and Axes of the ploted confusion matrix
        """
        return plot.confusion_matrix(self.confusion_matrix, ['P', 'N'], f'Confusion Matrix for \"{self.cls}\" Class')

    def __str__(self):
        no_instances_string = ' [NO INSTANCES]' if not self.has_instances else ''
        return (
            f'{self.cls}{no_instances_string}:\n'
            f'\tTP: {self.TP}\n'
            f'\tTN: {self.TN}\n'
            f'\tFP: {self.FP}\n'
            f'\tFN: {self.FN}\n'
            f'\tFPR: {self.false_positive_rate}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tRecall: {self.recall}\n'
            f'\tPrecision: {self.precision}\n'
            f'\tSpecificity: {self.specificity}\n'
            f'\tF-Score: {self.f_score}'
        )


class ClassificationMetrics:
    """Multiclass classification metrics

    Attributes:
        classes (List[str]): Class names.
        confusion_matrix (List[List[int]]): Confusion matrix of all the classes.
    """

    def __init__(self, classes: List[str], confusion_matrix: List[List[int]] = []):
        """
        Args:
            classes (List[str]): Class names.
            confusion_matrix (List[List[int]], optional): Confusion matrix of all the classes. Defaults to [].

        Raises:
            AssertException: If the confusion matrix is not a square matrix of order `len(classes)`
        """
        self._confusion_matrix = np.array(confusion_matrix)

        assert self._confusion_matrix.shape == (len(classes), len(
            classes)), 'The confusion matrix must be a square matrix of order len(classes)'

        self._classes = classes

        self._by_class = self.__get_by_class_metrics()
        self._by_class_w_instances = list(filter(
            lambda c: c.has_instances, self._by_class))
        self._count = self._confusion_matrix.sum()

    @property
    def by_class(self) -> List[BinaryClassificationMetrics]:
        """List[BinaryClassificationMetrics]: Binary metrics calculated for each class index."""
        return self._by_class

    @property
    def by_class_w_instances(self) -> List[BinaryClassificationMetrics]:
        """List[BinaryClassificationMetrics]: Binary metrics calculated for each class index with instances."""
        return self._by_class_w_instances

    @property
    def count(self) -> int:
        """int: Total count of classified instances."""
        return self._count

    @property
    def accuracy(self) -> float:
        """float: Accuracy metric - correct classifications / count."""
        return np.diagonal(self._confusion_matrix).sum() / self.count

    @property
    def avg_recall(self) -> float:
        """float: Macro average recall metric."""
        return reduce(lambda acc, curr: curr.recall + acc, self.by_class_w_instances, .0) / len(self.by_class_w_instances)

    @property
    def avg_precision(self) -> float:
        """float: Macro average precision metric."""
        return reduce(lambda acc, curr: curr.precision + acc, self.by_class_w_instances, .0) / len(self.by_class_w_instances)

    @property
    def avg_specificity(self) -> float:
        """float: Macro average specificity metric."""
        return reduce(lambda acc, curr: curr.specificity + acc, self.by_class_w_instances, .0) / len(self.by_class_w_instances)

    @property
    def avg_f_score(self) -> float:
        """float: Macro average F-Score/Dice metric."""
        return reduce(lambda acc, curr: curr.f_score + acc, self.by_class_w_instances, .0) / len(self.by_class_w_instances)

    @property
    def avg_false_positive_rate(self) -> float:
        """float: Macro average False Positive Rate metric."""
        return reduce(lambda acc, curr: curr.false_positive_rate + acc, self.by_class_w_instances, .0) / len(self.by_class_w_instances)

    @property
    def confusion_matrix(self) -> List[List[int]]:
        """List[List[int]]: Confusion matrix of all the classes"""
        return self._confusion_matrix

    def show_confusion_matrix(self):
        """Plots de confusion matrix

        Return:
            Tuple[Figure, Axes]: Figure and Axes of the ploted confusion matrix
        """
        return plot.confusion_matrix(self._confusion_matrix, self._classes)

    def __get_by_class_metrics(self):
        by_class = []
        for i in range(0, len(self._classes)):
            tp = self._confusion_matrix[i, i]
            tn = np.delete(np.delete(self._confusion_matrix, i, 0), i, 1).sum()
            fp = self._confusion_matrix[i, :].sum() - tp
            fn = self._confusion_matrix[:, i].sum() - tp

            by_class.append(BinaryClassificationMetrics(
                cls=self._classes[i],
                TP=tp,
                TN=tn,
                FP=fp,
                FN=fn
            ))
        return by_class

    def __str__(self):
        return (
            f'Classification Metrics:\n'
            f'\tSample Count: {self.count}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tAvg Recall: {self.avg_recall}\n'
            f'\tAvg Precision: {self.avg_precision}\n'
            f'\tAvg Specificity: {self.avg_specificity}\n'
            f'\tAvg FPR: {self.avg_false_positive_rate}\n'
            f'\tAvg F-Score: {self.avg_f_score}\n\n'
            f'By Class:\n\n'
        ) + '\n'.join([cls_metrics.__str__()
                       for cls_metrics in self.by_class])


class BinarySegmentationMetrics(BinaryClassificationMetrics):
    """Binary pixel-based classification metrics

    Attributes:
        TP (int): True Positive pixels count.
        FP (int): False Positive pixels count.
        TN (int): True Negative pixels count.
        FN (int): False Negative pixels count.
    """

    def __init__(self, classification_metrics: BinaryClassificationMetrics):
        """
        Args:
            classification_metrics (BinaryClassificationMetrics): Pixel-based binary classification metrics 
        """
        super().__init__(
            cls=classification_metrics.cls,
            TP=classification_metrics.TP,
            TN=classification_metrics.TN,
            FP=classification_metrics.FP,
            FN=classification_metrics.FN
        )

    @property
    def iou(self) -> float:
        """float: IoU/Jaccard Index metric - TP / (FP + FN + TP)."""
        return self.TP / (self.FP + self.FN + self.TP)

    def __str__(self):
        return (
            f'{self.cls}:\n'
            f'\tTP: {self.TP}\n'
            f'\tTN: {self.TN}\n'
            f'\tFP: {self.FP}\n'
            f'\tFN: {self.FN}\n'
            f'\tIoU: {self.iou}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tRecall: {self.recall}\n'
            f'\tPrecision: {self.precision}\n'
            f'\tSpecificity: {self.specificity}\n'
            f'\tFPR: {self.false_positive_rate}\n'
            f'\tF-Score: {self.f_score}\n'
        )


class SegmentationMetrics(ClassificationMetrics):
    """Multiclass pixel-based classification metrics

    Attributes:
        classes (List[str]): Class names with `classes[0]` as the background class.
        confusion_matrix (List[List[int]]): Confusion matrix of all the classes.
    """

    def __init__(self, classes: List[str], confusion_matrix: List[List[int]] = []):
        """
        Args:
            classes (List[str]): Class names. It is expected the first class to be the background class.
            confusion_matrix (List[List[int]], optional): Confusion matrix of all the classes. Defaults to [].

        Raises:
            AssertException: If the confusion matrix is not a square matrix of order `len(classes)`.
            AssertException: If the number of classes is less than 2.
        """
        assert len(classes) > 1, \
            'There should be at least two classes (with background as the first)'
        super().__init__(classes, confusion_matrix)
        self._by_class = [BinarySegmentationMetrics(x) for x in self.by_class]
        self._by_class_w_instances = [
            x for x in self.by_class if x.has_instances]

    @property
    def avg_iou(self) -> float:
        """float: Macro average IoU/Jaccard Index metric."""
        return reduce(lambda acc, curr: curr.iou + acc, self._by_class_w_instances, .0) / len(self._by_class_w_instances)

    @property
    def avg_iou_no_bkg(self) -> float:
        """float: Macro average IoU/Jaccard Index metric without `background` class (index 0)."""
        return reduce(lambda acc, curr: curr.iou + acc, self._by_class_w_instances[1:], .0) / (len(self._by_class_w_instances) - 1)

    def __str__(self):
        return (
            f'Segmentation Metrics:\n'
            f'\tPixel Count: {self.count}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tAvg Recall: {self.avg_recall}\n'
            f'\tAvg Precision: {self.avg_precision}\n'
            f'\tAvg Specificity: {self.avg_specificity}\n'
            f'\tAvg F-Score: {self.avg_f_score}\n'
            f'\tAvg FPR: {self.avg_false_positive_rate}\n'
            f'\tAvg IoU: {self.avg_iou}\n'
            f'\tAvg IoU w/o Background: {self.avg_iou_no_bkg}\n'
            f'By Class:\n\n'
        ) + '\n'.join(list([cls_metrics.__str__()
                            for cls_metrics in self.by_class]))


class BinaryDetectionMetrics(BinaryClassificationMetrics):
    """Binary instance-based classification metrics

    Attributes:
        TP (int): True Positive instances count.
        FP (int): False Positive instances count.
        TN (int): Not used in object detection. Always 0.
        FN (int): False Negative instances count.
    """

    def __init__(self, classification_metrics: BinaryClassificationMetrics,
                 iou: float,
                 predictions: List[PredictionResult]):
        super().__init__(
            cls=classification_metrics.cls,
            TP=classification_metrics.TP,
            FP=classification_metrics.FP,
            FN=classification_metrics.FN
        )
        self._iou = iou
        self._precision_recall_curve = self.__calculate_precision_recall_curve(
            predictions) if self.gt_count > 0 else []

    @property
    def gt_count(self) -> int:
        """int: Total count of GT bboxes."""
        return self.TP + self.FN

    @property
    def predicted_count(self) -> int:
        """int: Total count of predicted bboxes."""
        return self.TP + self.FP

    @property
    def iou(self) -> float:
        """float: IoU/Jaccard Index metric.

        This metric is calculated as for segmentation, considering bboxes as pixel masks.
        """
        return self._iou

    @property
    def precision_recall_curve(self) -> List[Tuple[float, float]]:
        """List[Tuple[float, float]]: Precision x Recall curve as a list of (Recall, Precision) tuples."""
        assert self.gt_count > 0, "This class does not have instances."
        return self._precision_recall_curve

    def average_precision(self, interpolation_points: Optional[int] = None) -> float:
        """Calculates the Average Precision metric.

        Args:
            interpolation_points (Optional[int], optional): Number of points to use for interpolation. 
            Uses all points if None. Defaults to None.

        Returns:
            float: Average Precision metric.
        """
        if not interpolation_points is None:
            return self._interpolated_average_precision(interpolation_points)
        else:
            return self._average_precision()

    def show_precision_recall_curve(self):
        """Plots de Precision x Recall curve.

        Return:
            Tuple[Figure, Axes]: Figure and Axes of the ploted Precision x Recall curve.
        """
        return plot.precision_recall_curve([self._precision_recall_curve], [self.cls])

    def __calculate_precision_recall_curve(self, predictions: List[PredictionResult]) -> List[Tuple[float, float]]:
        assert self.gt_count > 0, "The GT count must be greater than 0."

        sorted_predictions = sorted(
            predictions, key=lambda p: float(p.score), reverse=True)

        curve: List[Tuple[float, float]] = []
        tp = 0
        fp = 0
        gt_count = self.gt_count

        def calc_precision(tp, fp): return tp/(tp + fp)
        def calc_recall(tp): return tp/gt_count

        for prediction in sorted_predictions:
            if prediction.type == PredictionResultType.TP:
                tp += 1
            elif prediction.type == PredictionResultType.FP:
                fp += 1
            curve.append((calc_recall(tp), calc_precision(tp, fp)))

        return curve

    def _interpolated_average_precision(self, interpolation_points: int) -> float:
        precision_sum = .0
        recall_interpolation = np.linspace(0, 1, interpolation_points)

        for point in recall_interpolation:
            precision_sum += max([rp[1]
                                  for rp in self._precision_recall_curve if rp[0] >= point] + [0])

        return precision_sum / interpolation_points

    def _average_precision(self) -> float:
        ap_sum = .0
        prev_max_recall = .0

        greater_recall_points = self._precision_recall_curve
        while len(greater_recall_points) > 0:
            max_point = max(greater_recall_points, key=lambda rp: rp[1])
            ap_sum += (max_point[0] - prev_max_recall) * max_point[1]
            prev_max_recall = max_point[0]
            greater_recall_points = [
                rp for rp in greater_recall_points if rp[0] > prev_max_recall]

        return ap_sum

    def __str__(self):
        return (
            f'{self.cls}:\n'
            f'\tTP: {self.TP}\n'
            f'\tFP: {self.FP}\n'
            f'\tFN: {self.FN}\n'
            f'\tTN: [NA]\n'
            f'\tIoU: {self.iou}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tRecall: {self.recall}\n'
            f'\tPrecision: {self.precision}\n'
            f'\tF-Score: {self.f_score}\n'
            f'\tAverage Precision: {self.average_precision()}\n'
            f'\t11-point Average Precision: {self.average_precision(11)}\n'
        )


class DetectionMetrics(ClassificationMetrics):
    """Multiclass detection metrics

    Attributes:
        classes (List[str]): Class names. The last must be the "undetected" class.
        confusion_matrix (List[List[int]]): Confusion matrix of all the classes.
        iou_by_class (List[float]): IoU values indexed by class.

        The last column and line must correspond to the "undetected" class.
    """

    def __init__(self, classes: List[str],
                 confusion_matrix: List[List[int]],
                 iou_by_class: List[float],
                 predictions_by_class: List[List[PredictionResult]]):
        super().__init__(classes, confusion_matrix)
        by_class_wo_undetected = self.by_class[slice(0, -1)]
        self._by_class = [BinaryDetectionMetrics(by_class, iou, predictions)
                          for by_class, iou, predictions in zip(by_class_wo_undetected, iou_by_class, predictions_by_class)]
        self._by_class_w_instances = [
            x for x in self.by_class if x.has_instances]

    @property
    def avg_iou(self):
        """float: Macro average IoU/Jaccard Index metric.

        This metric is calculated as for segmentation, considering bboxes as pixel masks.
        """
        return reduce(lambda acc, curr: curr.iou + acc, self._by_class_w_instances, .0) / len(self._by_class_w_instances)

    def mean_average_precision(self, interpolation_points: Optional[int] = None) -> float:
        """Calculates the mean of Average Precision metrics of all classes.

        Args:
            interpolation_points (Optional[int], optional): Number of points to use for interpolation. 
            Uses all points if None. Defaults to None.

        Returns:
            float: Mean Average Precision metric.
        """
        return reduce(lambda acc, curr: curr.average_precision(interpolation_points) + acc, self._by_class_w_instances, .0) / len(self._by_class_w_instances)

    def __str__(self):
        return (
            f'Detection Metrics:\n'
            f'\tBboxes Count: {self.count}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tAvg Recall: {self.avg_recall}\n'
            f'\tAvg Precision: {self.avg_precision}\n'
            f'\tAvg F-Score: {self.avg_f_score}\n'
            f'\tAvg IoU: {self.avg_iou}\n'
            f'By Class:\n\n'
        ) + '\n'.join(list([cls_metrics.__str__()
                            for cls_metrics in self.by_class]))



from lapixdl.evaluation.model import BinaryDetectionMetrics, BinaryClassificationMetrics, PredictionResult, PredictionResultType
import pytest

# From https://github.com/rafaelpadilla/Object-Detection-Metrics
predictions = [
    PredictionResult(.88, PredictionResultType.FP),
    PredictionResult(.70, PredictionResultType.TP),
    PredictionResult(.80, PredictionResultType.FP),
    PredictionResult(.71, PredictionResultType.FP),
    PredictionResult(.54, PredictionResultType.TP),
    PredictionResult(.74, PredictionResultType.FP),
    PredictionResult(.18, PredictionResultType.TP),
    PredictionResult(.67, PredictionResultType.FP),
    PredictionResult(.38, PredictionResultType.FP),
    PredictionResult(.91, PredictionResultType.TP),
    PredictionResult(.44, PredictionResultType.FP),
    PredictionResult(.35, PredictionResultType.FP),
    PredictionResult(.78, PredictionResultType.FP),
    PredictionResult(.45, PredictionResultType.FP),
    PredictionResult(.14, PredictionResultType.FP),
    PredictionResult(.62, PredictionResultType.TP),
    PredictionResult(.44, PredictionResultType.FP),
    PredictionResult(.95, PredictionResultType.TP),
    PredictionResult(.23, PredictionResultType.FP),
    PredictionResult(.45, PredictionResultType.FP),
    PredictionResult(.84, PredictionResultType.FP),
    PredictionResult(.43, PredictionResultType.FP),
    PredictionResult(.48, PredictionResultType.TP),
    PredictionResult(.95, PredictionResultType.FP)
]

TP = len([prediction for prediction in predictions if prediction.type ==
          PredictionResultType.TP])
FP = len([prediction for prediction in predictions if prediction.type ==
          PredictionResultType.FP])
FN = 15 - TP


def test_gt_count():
    bin_class = BinaryClassificationMetrics(cls='a', FN=FN, TP=TP, FP=FP)
    metrics = BinaryDetectionMetrics(bin_class, 0, predictions)

    assert metrics.gt_count == 15


def test_pred_count():
    bin_class = BinaryClassificationMetrics(cls='a', FN=FN, TP=TP, FP=FP)
    metrics = BinaryDetectionMetrics(bin_class, 0, predictions)

    assert metrics.predicted_count == 24


def test_iou():
    bin_class = BinaryClassificationMetrics(cls='a', FN=FN, TP=TP, FP=FP)
    metrics = BinaryDetectionMetrics(bin_class, 10, predictions)

    assert metrics.iou == 10


def test_precision_recall_curve():
    bin_class = BinaryClassificationMetrics(cls='a', FN=FN, TP=TP, FP=FP)
    metrics = BinaryDetectionMetrics(bin_class, 10, predictions)

    rounded_pr_curve = [(round(rp[0], 4), round(rp[1], 4))
                        for rp in metrics.precision_recall_curve]
    expected_curve = [
        (0.0667, 1.0000),
        (0.0667, 0.5000),
        (0.1333, 0.6667),
        (0.1333, 0.5000),
        (0.1333, 0.4000),
        (0.1333, 0.3333),
        (0.1333, 0.2857),
        (0.1333, 0.2500),
        (0.1333, 0.2222),
        (0.2000, 0.3000),
        (0.2000, 0.2727),
        (0.2667, 0.3333),
        (0.3333, 0.3846),
        (0.4000, 0.4286),
        (0.4000, 0.4000),
        (0.4000, 0.3750),
        (0.4000, 0.3529),
        (0.4000, 0.3333),
        (0.4000, 0.3158),
        (0.4000, 0.3000),
        (0.4000, 0.2857),
        (0.4000, 0.2727),
        (0.4667, 0.3043),
        (0.4667, 0.2917),
    ]

    assert rounded_pr_curve == expected_curve

def test_average_precision():
    bin_class = BinaryClassificationMetrics(cls='a', FN=FN, TP=TP, FP=FP)
    metrics = BinaryDetectionMetrics(bin_class, 10, predictions)

    assert round(metrics.average_precision(11), 4) == .2684
    assert round(metrics.average_precision(), 4) == .2457



from typing import List, Tuple, Optional

import pandas as pd
import seaborn as sn
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes


def confusion_matrix(confusion_matrix: List[List[int]], classes: List[str], title: Optional[str] = "Confusion Matrix") -> Tuple[Figure, Axes]:
    """Plots a confusion matrix

    Args:
        confusion_matrix (List[List[int]]): Confusion matrix.
        classes (List[str]): Class names.
        title (Optional[str], optional): Chart title. Defaults to "Confusion Matrix".

    Return:
        Tuple[Figure, Axes]: Figure and Axes of the ploted confusion matrix.
    """

    df_cm = pd.DataFrame(confusion_matrix, classes, classes)

    fig, ax = plt.subplots()
    fig.suptitle(title)

    ax = sn.heatmap(df_cm,
                    annot=True,
                    ax=ax,
                    cmap="YlGnBu")

    ax.set_xlabel("Actual")
    ax.set_ylabel("Predicted")
    plt.show()

    return fig, ax


def precision_recall_curve(recalls_precisions_by_class: List[List[Tuple[float, float]]],
                           classes: List[str],
                           title: Optional[str] = "Precision x Recall Curve") -> Tuple[Figure, Axes]:
    """Plots a Precision x Recall curve.

    Args:
        recalls_precisions_by_class (List[List[Tuple[float, float]]]): (Recall, Precision) tuples indexed by class.
        classes (List[str]): Class names.
        title (Optional[str], optional): Chart title. Defaults to "Precision x Recall Curve".

    Return:
        Tuple[Figure, Axes]: Figure and Axes of the ploted Precision x Recall curve.
    """

    table = []
    for i, recalls_precisions_cls in enumerate(recalls_precisions_by_class):
        for recall_precision in recalls_precisions_cls:
            table.append(
                (classes[i], recall_precision[0], recall_precision[1]))

    df = pd.DataFrame(table, columns=["class", "recall", "precision"])

    fig, ax = plt.subplots()
    fig.suptitle(title)

    ax = sn.lineplot(data=df,
                     y="precision",
                     x="recall",
                     hue="class",
                     ax=ax)

    ax.set_xlabel("Recall")
    ax.set_ylabel("Precision")
    plt.show()

    return fig, ax



from typing import List, Optional, Tuple
import numpy as np
import itertools


def __bounding_rect_from_polygon(polygon: List[Tuple[int, int]]) -> List[int]:
    arr = np.array(polygon)

    max = arr.max(axis=0)
    min = arr.min(axis=0)
    h = max[1] - min[1]
    w = max[0] - min[0]

    return [min[0], min[1], w, h]


def __generate_coco_annotations(img_labels: List[dict], img_id: int, annotation_first_id: int, classes: List[str]) -> List[dict]:
    annotations = []

    for img_label in img_labels:
        class_name = img_label['value']
        polygon = img_label['polygon']
        poly_convert = [(p['x'], p['y']) for p in polygon]
        bbox = __bounding_rect_from_polygon(poly_convert)
        segmentation = list(itertools.chain.from_iterable(poly_convert))
        if class_name not in classes:
            classes.append(class_name)

        annotations.append({
            'segmentation': [segmentation],
            'iscrowd': 0,
            'image_id': img_id,
            'category_id': classes.index(class_name),
            'id': annotation_first_id,
            'bbox': bbox,
            'area': __calculate_area(segmentation)
        })

        annotation_first_id += 1

    return annotations


def __calculate_area(segmentation: List[float]) -> float:
    x = segmentation[0:-1:2]
    y = segmentation[1::2]
    return 0.5*np.abs(np.dot(x, np.roll(y, 1))-np.dot(y, np.roll(x, 1))) #Shoelace


def __generate_coco_file(lblbox_annotations: dict, img_names_to_include: Optional[List[str]] = None) -> dict:
    annotation_id = 1
    image_id = 1

    annotations: List[dict] = []
    images: List[dict] = []
    classes: List[str] = []

    ftr_lblbox_annotations = (filter(lambda ant: ant['External ID'] in (img_names_to_include or []),
                                     lblbox_annotations)) if img_names_to_include is not None else lblbox_annotations

    for images_labels in ftr_lblbox_annotations:
        img_filename = images_labels['External ID']
        img_labels = images_labels['Label']['objects'] if 'objects' in images_labels['Label'] else [
        ]

        new_annotations = __generate_coco_annotations(
            img_labels, image_id, annotation_id, classes)
        annotations += new_annotations

        images.append({
            "file_name": img_filename,
            "height": 1200,
            "width": 1600,
            "id": image_id
        })

        image_id += 1
        annotation_id += len(new_annotations)

    categories = [{
        "supercategory": "none",
        "name": cls,
        "id": i
    } for i, cls in enumerate(classes)]

    final_json = {
        "type": "instances",
        "images": images,
        "categories": categories,
        "annotations": annotations
    }

    return final_json


def labelbox_to_coco(lblbox_annotations: dict, img_names_to_include: Optional[List[str]] = None) -> dict:
    """Converts from Labelbox format to the COCO format

    Args:
        lblbox_annotations (dict): Dict of labelbox annotations parsed from the exported json.
        img_names_to_include (Optional[List[str]], optional): Image names (External IDs) to include in the converted file. Defaults to None.

    Returns:
        dict: Dict in the COCO format.
    """
    return __generate_coco_file(lblbox_annotations, img_names_to_include)



from typing import List, Iterable

import numpy as np
from tqdm import tqdm

from .model import *


def evaluate_segmentation(gt_masks: Iterable[Mask],
                          pred_masks: Iterable[Mask],
                          classes: List[str]) -> SegmentationMetrics:
    """Evaluates segmentation predictions

    The iterables should return one mask per iteration and the itens of
    `gt_masks` and `pred_masks` with same index should correspond to the
    same sample.

    Masks should be 2D arrays where each value corresponds to the class 
    index of the pixel the sample image.

    Args:
        gt_masks (Iterable[Mask]): Iterable of ground truth masks.
        pred_masks (Iterable[Mask]): Iterable of predicted masks.
        classes (List[str]): Class names.

    Returns:
        SegmentationMetrics: Pixel-based classification and segmentation metrics.
    """

    confusion_matrix = np.zeros((len(classes), len(classes)), np.int)

    for (curr_gt_mask, curr_pred_mask) in tqdm(zip(gt_masks, pred_masks), unit=' masks'):
        flat_gt = __flat_mask(curr_gt_mask)
        flat_pred = __flat_mask(curr_pred_mask)
        for (curr_pred, curr_gt) in zip(flat_pred, flat_gt):
            confusion_matrix[curr_pred, curr_gt] += 1

    metrics = SegmentationMetrics(classes, confusion_matrix)

    print(metrics)

    return metrics


def evaluate_classification(gt_classifications: Iterable[Classification],
                            pred_classifications: Iterable[Classification],
                            classes: List[str]) -> ClassificationMetrics:
    """Evaluates classification predictions

    The iterables should return one classification per iteration and 
    the itens of `gt_classifications` and `pred_classifications` with 
    same index should correspond to the same sample.

    Args:
        gt_classifications (Iterable[Classification]): Ground truth classifications.
        pred_classifications (Iterable[Classification]): Predicted classifications.
        classes (List[str]): Class names.

    Returns:
        ClassificationMetrics: Classification metrics.
    """

    confusion_matrix = np.zeros((len(classes), len(classes)), np.int)

    for (curr_gt_classification, curr_pred_classification) in tqdm(zip(gt_classifications, pred_classifications), unit=' samples'):
        confusion_matrix[curr_pred_classification.cls,
                         curr_gt_classification.cls] += 1

    metrics = ClassificationMetrics(classes, confusion_matrix)

    print(metrics)

    return metrics


def evaluate_detection(gt_bboxes: Iterable[List[BBox]],
                       pred_bboxes: Iterable[List[BBox]],
                       classes: List[str],
                       iou_threshold: float = .5,
                       undetected_cls_name: str = "_undetected_") -> DetectionMetrics:
    """Evaluates detection predictions

    The iterables should return the bboxes of one image per iteration 
    and the itens of `gt_bboxes` and `pred_bboxes` with same index 
    should correspond to the same image.

    Args:
        gt_bboxes (Iterable[List[BBox]]): Ground truth classifications.
        pred_bboxes (Iterable[List[BBox]]): Predicted classifications.
        classes (List[str]): Class names.
        iou_threshold (float): Minimum IoU threshold to consider a detection as a True Positive.
        undetected_cls_name (str): Name to be used for undetected instances class. Defaults to "_undetected_".

    Returns:
        DetectionMetrics: Detection metrics.
    """
    tot_images = 0
    classes_count = len(classes)

    cls_ious_sum = np.zeros(classes_count)
    confusion_matrix = np.zeros((classes_count + 1, classes_count + 1), np.int)
    undetected_idx = classes_count
    # Tracks detection scores to calculate the Precision x Recall curve and the Average Precision metric
    predictions_by_class: List[List[PredictionResult]] = [
        [] for i in range(len(classes))]

    # For each image GT and predicted bbox set
    for (curr_gt_bboxes, curr_pred_bboxes) in tqdm(zip(gt_bboxes, pred_bboxes), unit=' samples'):
        pairwise_bbox_ious = calculate_pairwise_bbox_ious(
            curr_gt_bboxes, curr_pred_bboxes)

        # Sums classes' IoUs by image to calculate the average at the end
        curr_img_cls_ious = calculate_iou_by_class(
            curr_gt_bboxes, curr_pred_bboxes, classes_count)
        cls_ious_sum = np.add(cls_ious_sum, curr_img_cls_ious)
        tot_images += 1

        # Tracks prediction bboxes that does not correspond to any GT bbox to identify FPs
        no_hit_idxs = set(range(0, len(curr_pred_bboxes)))

        # Iterate through lines of predictions bbox IoUs for each GT bbox
        for i, gt_ious in enumerate(pairwise_bbox_ious):
            gt_cls_idx = curr_gt_bboxes[i].cls

            # Removes already matched predictions
            gt_ious = [0 if i not in no_hit_idxs else iou for i,
                       iou in enumerate(gt_ious)]

            if len(gt_ious) > 0:
                max_iou_idx = np.argmax(gt_ious)
                # Only the max IoU is considered TP, the others are FPs
                max_iou = gt_ious[max_iou_idx]
            else:
                max_iou = 0

            if max_iou < iou_threshold:  # FN - GT bbox not detected
                confusion_matrix[undetected_idx, gt_cls_idx] += 1
            else:  # TP - GT bbox detected
                no_hit_idxs.remove(max_iou_idx)  # Remove from FPs

                pred_cls_idx = curr_pred_bboxes[max_iou_idx].cls
                confusion_matrix[pred_cls_idx, gt_cls_idx] += 1
                if pred_cls_idx == gt_cls_idx:
                    predictions_by_class[pred_cls_idx]\
                        .append(PredictionResult(curr_pred_bboxes[max_iou_idx].score, PredictionResultType.TP))
                else:
                    predictions_by_class[pred_cls_idx]\
                        .append(PredictionResult(curr_pred_bboxes[max_iou_idx].score, PredictionResultType.FP))

        for no_hit_idx in no_hit_idxs:  # FPs - Predictions that do not match any GT
            pred_cls_idx = curr_pred_bboxes[no_hit_idx].cls
            confusion_matrix[pred_cls_idx, undetected_idx] += 1
            predictions_by_class[pred_cls_idx]\
                .append(PredictionResult(curr_pred_bboxes[no_hit_idx].score, PredictionResultType.FP))

    metrics = DetectionMetrics(
        classes + [undetected_cls_name],
        confusion_matrix,
        cls_ious_sum / tot_images,
        predictions_by_class)

    print(metrics)

    return metrics


def calculate_pairwise_bbox_ious(gt_bboxes: List[BBox],
                                 pred_bboxes: List[BBox]) -> List[List[float]]:
    """Calculates the [gt x pred] matrix of pairwise IoUs of GT and predicted bboxes of an image.

    Args:
        gt_bboxes (List[BBox]): GT bboxes
        pred_bboxes (List[BBox]): Predicted bboxes

    Returns:
        List[List[float]]: [gt x pred] matrix of pairwise IoUs of GT and predicted bboxes
    """

    ious = np.zeros((len(gt_bboxes), len(pred_bboxes)), np.float)

    for i, gt_bbox in enumerate(gt_bboxes):
        for j, pred_bbox in enumerate(pred_bboxes):
            ious[i, j] = calculate_bbox_iou(gt_bbox, pred_bbox)

    return ious


def calculate_bbox_iou(bbox_a: BBox, bbox_b: BBox) -> float:
    """Calculates the IoU between 2 bboxes.

    Args:
        bbox_a (BBox): BBox.
        bbox_b (BBox): BBox.

    Returns:
        float: IoU between the two bboxes.
    """

    intersect_area = bbox_a.intersection_area_with(bbox_b)
    union_area = bbox_a.union_area_with(bbox_b, intersect_area)

    return float(intersect_area) / float(union_area)


def calculate_iou_by_class(gt_bboxes: List[BBox],
                           pred_bboxes: List[BBox],
                           classes_count: int) -> List[float]:
    """Calculates the array of IoUs between GT and predicted bboxes of an image by class.
    This method considers bbox as segmentation masks to calculate the IoUs.

    Args:
        gt_bboxes (List[BBox]): GT bboxes
        pred_bboxes (List[BBox]): Predicted bboxes
        classes_count (int): Classes count

    Returns:
        List[float]: IoUs of an image indexed by class
    """

    ious = np.zeros(classes_count, np.float)

    for i in range(classes_count):
        ious[i] = __calculate_binary_iou([gt_bbox for gt_bbox in gt_bboxes if gt_bbox.cls == i],
                                         [pred_bbox for pred_bbox in pred_bboxes if pred_bbox.cls == i])

    return ious


def __calculate_binary_iou(gt_bboxes: List[BBox],
                           pred_bboxes: List[BBox]) -> float:
    if len(gt_bboxes) == 0 or len(pred_bboxes) == 0:
        return .0

    bboxes_btm_right_points = [bbox.bottom_right_point for bbox in gt_bboxes]\
        + [bbox.bottom_right_point for bbox in pred_bboxes]
    max_x = max([point[0] for point in bboxes_btm_right_points])
    max_y = max([point[1] for point in bboxes_btm_right_points])

    gt_mask = __draw_bboxes((int(max_x + 1), int(max_y + 1)), gt_bboxes)
    pred_mask = __draw_bboxes((int(max_x + 1), int(max_y + 1)), pred_bboxes)

    tp, f = (0, 0)
    flat_gt = __flat_mask(gt_mask)
    flat_pred = __flat_mask(pred_mask)
    for (curr_pred, curr_gt) in zip(flat_pred, flat_gt):
        if curr_gt == 1 and curr_pred == 1:
            tp += 1
        elif curr_gt != curr_pred:
            f += 1

    return tp / (f + tp)


def __draw_bboxes(mask_shape: Tuple[int, int], bboxes: List[BBox]) -> List[List[int]]:
    mask = np.zeros(mask_shape, np.int)

    for bbox in bboxes:
        mask[
            bbox.upper_left_point[0]:bbox.bottom_right_point[0] + 1,
            bbox.upper_left_point[1]:bbox.bottom_right_point[1] + 1
        ] = 1

    return mask


def __flat_mask(mask: Mask) -> List[int]:
    return [item for sublist in mask for item in sublist]



import pytest
import numpy as np

from lapixdl.evaluation.evaluate import evaluate_segmentation
from lapixdl.evaluation.model import BBox
from ... import utils


def test_evaluation_classification_metrics():
    classes = ['a', 'b', 'c']

    gt_masks = [
        [
            [0, 0, 0, 0, 0],
            [0, 1, 1, 1, 1],
            [1, 1, 1, 1, 1],
            [1, 2, 2, 2, 2],
            [2, 2, 2, 2, 2]
        ]
    ]
    pred_masks = [
        [
            [0, 0, 0, 0, 2],
            [1, 0, 0, 0, 0],
            [0, 0, 2, 2, 1],
            [1, 0, 0, 0, 2],
            [2, 2, 2, 2, 2]
        ]
    ]

    metrics = evaluate_segmentation(gt_masks, pred_masks, classes)

    assert metrics.count == 25
    assert round(metrics.accuracy, 3) == .48
    assert round(metrics.avg_recall, 3) == .511
    assert round(metrics.avg_precision, 3) == .547
    assert round(metrics.avg_f_score, 3) == .465
    assert round(metrics.avg_specificity, 3) == .757


def test_evalution_iou_with_image():
    classes = ['bkg', 'kite']

    mask_shape = (480, 640)

    gt_bbox = BBox(110, 110, 320, 280, 1)
    mask_bin_GT = utils.bin_mask_from_bb(mask_shape, gt_bbox)

    pred_bbox = BBox(70, 50, 240, 220, 1)
    mask_bin_pred = utils.bin_mask_from_bb(mask_shape, pred_bbox)

    metrics = evaluate_segmentation([mask_bin_GT], [mask_bin_pred], classes)

    assert metrics.count == 480 * 640
    assert round(metrics.by_class[1].iou, 3) == .290
    assert round(metrics.avg_iou_no_bkg, 3) == .290
    assert round(metrics.avg_iou, 3) == .502


def test_should_not_accept_single_class():
    classes_empty = []
    classes_bkg_only = ['bkg']

    with pytest.raises(AssertionError):
        evaluate_segmentation([], [], classes_empty)

    with pytest.raises(AssertionError):
        evaluate_segmentation([], [], classes_bkg_only)


def test_evalution_iou_multiclass_with_images():
    classes = ['bkg', 'kite', 'person']

    mask_shape = (480, 640)

    gt_bbox_1 = BBox(10, 10, 10, 10, 1)
    mask_bin_GT_1 = utils.bin_mask_from_bb(mask_shape, gt_bbox_1)

    pred_bbox_1 = BBox(10, 10, 10, 10, 1)
    mask_bin_pred_1 = utils.bin_mask_from_bb(mask_shape, pred_bbox_1)

    gt_bbox_2 = BBox(110, 110, 320, 280, 2)
    mask_bin_GT_2 = utils.bin_mask_from_bb(mask_shape, gt_bbox_2)

    pred_bbox_2 = BBox(70, 50, 240, 220, 2)
    mask_bin_pred_2 = utils.bin_mask_from_bb(mask_shape, pred_bbox_2)

    mask_multi_GT = np.maximum(mask_bin_GT_1, mask_bin_GT_2 * 2)
    mask_multi_pred = np.maximum(mask_bin_pred_1, mask_bin_pred_2 * 2)

    metrics = evaluate_segmentation([mask_multi_GT], [mask_multi_pred], classes)

    assert metrics.count == 480 * 640

    assert metrics.by_class[1].iou == 1.0
    assert round(metrics.by_class[2].iou, 3) == .290
    assert round(metrics.avg_iou_no_bkg, 3) == .645



from lapixdl.evaluation.evaluate import evaluate_classification
from lapixdl.evaluation.model import Classification

def test_evaluation():
    classes = ['a', 'b', 'c']

    gt_class = [Classification(x) for x in [0,0,0,0,0,0, 1,1,1,1,1,1,1,1,1,1, 2,2,2,2,2,2,2,2,2] ]
    pred_class = [Classification(x) for x in [0,0,0,0,2,1, 0,0,0,0,0,0,2,2,1,1, 0,0,0,2,2,2,2,2,2] ]


    metrics = evaluate_classification(gt_class, pred_class, classes)

    assert metrics.count == 25
    assert round(metrics.accuracy, 3) == .48
    assert round(metrics.avg_recall, 3) == .511
    assert round(metrics.avg_precision, 3) == .547
    assert round(metrics.avg_f_score, 3) == .465
    assert round(metrics.avg_specificity, 3) == .757


from lapixdl.evaluation.model import ClassificationMetrics, BinaryClassificationMetrics
import pytest


def test_constructor_invalid():
    with pytest.raises(AssertionError):
        ClassificationMetrics(['a', 'b'], [[1],[2]])

def test_constructor_valid():
    ClassificationMetrics(['a', 'b'], [[1,2],[2,1]])

def test_count():
    metrics = ClassificationMetrics(['a', 'b', 'c'], 
    [
        [4,6,3],
        [1,2,0],
        [1,2,6]
    ])

    assert metrics.count == 25

    class_a = metrics.by_class[0]
    class_b = metrics.by_class[1]
    class_c = metrics.by_class[2]

    assert class_a.count == 25
    assert class_b.count == 25
    assert class_c.count == 25


def test_accuracy():
    metrics = ClassificationMetrics(['a', 'b', 'c'], 
    [
        [4,6,3],
        [1,2,0],
        [1,2,6]
    ])

    assert round(metrics.accuracy, 3) == .480

    class_a = metrics.by_class[0]
    class_b = metrics.by_class[1]
    class_c = metrics.by_class[2]

    assert round(class_a.accuracy, 3) == .560
    assert round(class_b.accuracy, 3) == .640
    assert round(class_c.accuracy, 3) == .760


def test_recall():
    metrics = ClassificationMetrics(['a', 'b', 'c'], 
    [
        [4,6,3],
        [1,2,0],
        [1,2,6]
    ])

    assert round(metrics.avg_recall, 3) == .511

    class_a = metrics.by_class[0]
    class_b = metrics.by_class[1]
    class_c = metrics.by_class[2]

    assert round(class_a.recall, 3) == .667
    assert round(class_b.recall, 3) == .200
    assert round(class_c.recall, 3) == .667



def test_specificity():
    metrics = ClassificationMetrics(['a', 'b', 'c'], 
    [
        [4,6,3],
        [1,2,0],
        [1,2,6]
    ])

    assert round(metrics.avg_specificity, 3) == .757

    class_a = metrics.by_class[0]
    class_b = metrics.by_class[1]
    class_c = metrics.by_class[2]

    assert round(class_a.specificity, 3) == .526
    assert round(class_b.specificity, 3) == .933
    assert round(class_c.specificity, 3) == .812


def test_precision():
    metrics = ClassificationMetrics(['a', 'b', 'c'], 
    [
        [4,6,3],
        [1,2,0],
        [1,2,6]
    ])

    assert round(metrics.avg_precision, 3) == .547

    class_a = metrics.by_class[0]
    class_b = metrics.by_class[1]
    class_c = metrics.by_class[2]

    assert round(class_a.precision, 3) == .308
    assert round(class_b.precision, 3) == .667
    assert round(class_c.precision, 3) == .667


def test_f_score():
    metrics = ClassificationMetrics(['a', 'b', 'c'], 
    [
        [4,6,3],
        [1,2,0],
        [1,2,6]
    ])

    assert round(metrics.avg_f_score, 3) == .465

    class_a = metrics.by_class[0]
    class_b = metrics.by_class[1]
    class_c = metrics.by_class[2]

    assert round(class_a.f_score, 3) == .421
    assert round(class_b.f_score, 3) == .308
    assert round(class_c.f_score, 3) == .667



from lapixdl.convert.labelbox import labelbox_to_coco, __calculate_area
import json

def test_labelbox_to_coco():
    with open('./tests/convert/labelbox_test.json') as file:
        labelbox_file = json.load(file)

    with open('./tests/convert/coco_expect.json') as file:
        coco_expect = json.load(file)

    conversion = labelbox_to_coco(labelbox_file)

    assert conversion == coco_expect

def test_labelbox_to_coco_w_filter():
    imgs_to_include = [
        "2019_07_10__16_23__0048_b0s0c0x135537-1600y50862-1200m6486.png",
        "2019_07_10__16_23__0048_b0s0c0x123913-1600y119932-1200m15182.png"
    ]

    with open('./tests/convert/labelbox_test.json') as file:
        labelbox_file = json.load(file)

    with open('./tests/convert/coco_expect_filtered.json') as file:
        coco_expect = json.load(file)

    conversion = labelbox_to_coco(labelbox_file, imgs_to_include)

    assert conversion == coco_expect

def test_calculate_area():
    segmentation = [0,0, 0,2, 2,2, 2,0]
    area = __calculate_area(segmentation)

    assert area == 4


from typing import Optional, Union, List, Tuple, TypeVar, Generic
import math

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib import cm
import matplotlib.patches as mpatches
import seaborn as sn
import numpy as np

from .model import BBox, Mask, Classification, Result, Image

correct_color = sn.color_palette("Paired")[3]
incorrect_color = sn.color_palette("Paired")[5]


def show_classifications(
        results: List[Result[Classification]],
        class_names: List[str],
        cols: int = 3,
        diff_correct_incorect: bool = True) -> Tuple[Figure, Axes]:
    """Shows multiple classification results.

    Args:
        results (List[Result[Classification]]): List of classification results.
        class_names (List[str]): Class names.
        cols (int, optional): Number of colunms to show. Defaults to 3.
        diff_correct_incorect (bool, optional): Indicates if correct and incorrect 
        results should be differentiated by color. Defaults to True.

    Returns:
        Tuple[Figure, Axes]: Figure and Axes of the ploted results.
    """

    rows = math.ceil(len(results) / cols)
    fig, axes = plt.subplots(rows, cols)
    fig.suptitle('Classifications')

    for i, result in enumerate(results):
        axe = axes[i // cols][i % cols]
        axe.set_title(
            f'GT: {class_names[result.gt.cls]}' +
            (f'\nPred: {class_names[result.prediction.cls]} ({result.prediction.score})' if not result.prediction is None else ''),
            fontsize='small',
            color='#333' if result.prediction is None or not diff_correct_incorect
            else (correct_color if result.prediction.cls ==
                  result.gt.cls else incorrect_color)
        )
        axe.axis('off')
        axe.imshow(result.image)

    for i in range(len(results), (cols * rows)):
        axe = axes[i // cols][i % cols]
        axe.axis('off')

    plt.tight_layout(w_pad=.2, h_pad=1.5)
    plt.show()
    return fig, axes


def show_segmentations(
        results: List[Result[Mask]],
        class_names: List[str],
        cmap: Optional[str] = 'tab10',
        mask_alpha: float = .3) -> Tuple[Figure, Axes]:
    """Shows segmentation results

    Args:
        results (List[Result[Mask]]): Segmentation results.
        class_names (List[str]): Class names.
        cmap (Optional[str], optional): Matplotlib color map for the mask. Defaults to 'tab10'.
        mask_alpha (float, optional): Alpha value for the mask colors. Defaults to .3.

    Returns:
        Tuple[Figure, Axes]: Figure and Axes of the ploted results.
    """

    cmap_colors = cm.get_cmap(cmap).colors

    rows = len(results)
    fig, axes = plt.subplots(rows, 3)

    fig.suptitle(' ', fontsize=40)  # To keep space for the legend
    lengend_handles = [mpatches.Patch(color=cmap_colors[i], label=code)
                       for i, code in enumerate(class_names)]
    fig.legend(handles=lengend_handles, fontsize='small',
               ncol=len(class_names), loc='upper center')

    for i, result in enumerate(results):
        axe_img = axes[i][0]
        axe_GT = axes[i][1]
        axe_pred = axes[i][2]

        axe_img.set_title('Image', fontsize='small')
        axe_img.imshow(result.image)
        axe_img.axis('off')

        axe_GT.set_title('GT', fontsize='small')
        axe_GT.axis('off')
        axe_GT.imshow(result.image)
        axe_GT.imshow(result.gt, cmap=cmap, alpha=mask_alpha,
                      interpolation='none', vmin=0, vmax=len(cmap_colors) - 1)

        axe_pred.axis('off')
        if not result.prediction is None:
            axe_pred.set_title('Prediction', fontsize='small')
            axe_pred.imshow(result.image)
            axe_pred.imshow(result.prediction, cmap=cmap, alpha=mask_alpha,
                            interpolation='none', vmin=0, vmax=len(cmap_colors) - 1)

    plt.tight_layout(w_pad=.2, h_pad=3)
    plt.show()

    return fig, axes


def show_detections(results: List[Result[List[BBox]]],
                    class_names: List[str],
                    cmap: Optional[str] = 'tab10',
                    show_bbox_label: bool = True) -> Tuple[Figure, Axes]:
    """Shows detection results.

    Args:
        results (List[Result[List[BBox]]]): Detection results.
        class_names (List[str]): Class names.
        cmap (Optional[str], optional): Matplotlib color map for the bboxes. Defaults to 'tab10'.
        show_bbox_label (bool, optional): Indicates if the class label should be shown for each bbox. Defaults to True.

    Returns:
        Tuple[Figure, Axes]: Figure and Axes of the ploted results.
    """

    cmap_colors = cm.get_cmap(cmap).colors

    rows = len(results)
    fig, axes = plt.subplots(rows, 3)

    fig.suptitle(' ', fontsize=40)  # To keep space for the legend
    lengend_handles = [mpatches.Patch(color=cmap_colors[i], label=code)
                       for i, code in enumerate(class_names)]
    fig.legend(handles=lengend_handles, fontsize='small',
               ncol=len(class_names), loc='upper center')

    for i, result in enumerate(results):
        axe_img = axes[i][0]
        axe_GT = axes[i][1]
        axe_pred = axes[i][2]

        axe_img.set_title('Image', fontsize='small')
        axe_img.imshow(result.image)
        axe_img.axis('off')

        axe_GT.set_title('GT', fontsize='small')
        axe_GT.axis('off')
        axe_GT.imshow(result.image)
        __draw_bboxes(axe_GT, result.gt, cmap_colors,
                      class_names, show_bbox_label)

        axe_pred.axis('off')
        if not result.prediction is None:
            axe_pred.set_title('Prediction', fontsize='small')
            axe_pred.imshow(result.image)
            __draw_bboxes(axe_pred, result.prediction,
                          cmap_colors, class_names, show_bbox_label)

    plt.tight_layout(w_pad=.2, h_pad=3)
    plt.show()

    return fig, axes


def __draw_bboxes(axe: plt.Axes, bboxes: List[BBox], cmap_colors: List, class_names: List[str], show_bbox_label: bool):
    for bbox in bboxes:
        color = cmap_colors[bbox.cls]
        label = (f'{class_names[bbox.cls]}' if show_bbox_label else '') + \
            (f' ({round(bbox.score, 3)})' if bbox.score else '')
        rect = mpatches.Rectangle(bbox.upper_left_point,
                                  bbox.width, bbox.height,
                                  linewidth=2,
                                  edgecolor=color,
                                  facecolor='none')
        axe.add_patch(rect)
        axe.annotate(label.strip(),
                     (bbox.upper_left_point[0] + 5,
                      bbox.upper_left_point[1] - 5),
                     color='w',
                     weight='bold',
                     fontsize=10,
                     ha='left', va='bottom',
                     bbox=dict(facecolor=color, edgecolor='none', pad=1.5))



import pytest
import numpy as np

from lapixdl.evaluation.evaluate import evaluate_detection
from lapixdl.evaluation.model import BBox
from ... import utils


def test_evaluation_detection_iou_metric():
    classes = ['kite']

    gt_bbox = BBox(110, 110, 320, 280, 0)

    pred_bbox = BBox(70, 50, 240, 220, 0, .5)

    metrics = evaluate_detection([[gt_bbox]], [[pred_bbox]], classes)

    assert round(metrics.avg_iou, 3) == .290
    assert round(metrics.by_class[0].iou, 3) == .290

def test_evaluation_detection_iou_metric_w_more_classes():
    classes = ['kite', 'person']

    gt_bbox = BBox(110, 110, 320, 280, 0)

    pred_bbox = BBox(70, 50, 240, 220, 0, .5)

    metrics = evaluate_detection([[gt_bbox]], [[pred_bbox]], classes)

    assert round(metrics.avg_iou, 3) == .290
    assert round(metrics.by_class[0].iou, 3) == .290

def test_evaluation_detection_no_gt():
    classes = ['kite', 'person']

    pred_bbox = BBox(70, 50, 240, 220, 0)

    metrics = evaluate_detection([[]], [[pred_bbox]], classes)

    assert round(metrics.avg_iou, 3) == 0
    assert round(metrics.by_class[0].iou, 3) == 0
    assert metrics._by_class[0].FP == 1 

def test_evaluation_detection_no_pred():
    classes = ['kite', 'person']

    gt_bbox = BBox(110, 110, 320, 280, 0)

    metrics = evaluate_detection([[gt_bbox]], [[]], classes)

    assert round(metrics.avg_iou, 3) == 0
    assert round(metrics.by_class[0].iou, 3) == 0
    assert metrics._by_class[0].FN == 1 



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='lapixdl',
    packages=find_packages(exclude=['tests']),
    version='0.7.10',
    description='Utils for Computer Vision Deep Learning research',

    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lapix-ufsc/lapixdl",

    author='LAPiX',
    license='MIT',

    install_requires=['numpy', 'tqdm', 'seaborn', 'pandas', 'matplotlib'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'cv2'],
    test_suite='tests',

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from lapixdl.evaluation.model import BinaryClassificationMetrics


def test_count():
    bin_class = BinaryClassificationMetrics(cls=['a', 'b'])
    bin_class.FN = 1
    bin_class.TP = 2

    assert bin_class.count == 3

    bin_class.FP = 2

    assert bin_class.count == 5


def test_accuracy():
    bin_class = BinaryClassificationMetrics(cls=['a', 'b'], FN=2, TN=2, TP=2, FP=2)

    assert bin_class.accuracy == 0.5


def test_recall():
    bin_class = BinaryClassificationMetrics(cls=['a', 'b'])
    bin_class.TP = 2
    bin_class.FN = 6

    assert bin_class.recall == 0.25


def test_fpr():
    bin_class = BinaryClassificationMetrics(cls=['a', 'b'])
    bin_class.FP = 2
    bin_class.TN = 6

    assert bin_class.false_positive_rate == 0.25

def test_specificity():
    bin_class = BinaryClassificationMetrics(cls=['a', 'b'])
    bin_class.TN = 6
    bin_class.FP = 2

    assert bin_class.specificity == 0.75


def test_precision():
    bin_class = BinaryClassificationMetrics(cls=['a', 'b'])
    bin_class.TP = 8
    bin_class.FP = 2

    assert bin_class.precision == 0.8


def test_f_score():
    bin_class = BinaryClassificationMetrics(cls=['a', 'b'])
    bin_class.FN = 4
    bin_class.TN = 2
    bin_class.TP = 5
    bin_class.FP = 6

    assert bin_class.f_score == 0.5
